

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h>

#define GLFW_DLL
#include <GL/glfw.h>
#include <GL/GL.h>
#include <GL/GLU.h>

#include "stb_image.c"

#define MAX_IMAGE_WIDTH 2048
#define MAX_IMAGE_HEIGHT 2048

#define MAX_WINDOW_WIDTH 1024
#define MAX_WINDOW_HEIGHT 768

#define TEXTURE_COUNT 3

#define LEVEL_MAX 4

#define PROGRAM_COUNT 2

#define FRAMEBUFFER_COUNT (LEVEL_MAX + 1)

/* Load a file. */
int LoadFile(const char* filename, char** data, size_t *dataSize)
{
	size_t nRead;
	char *tmp;
	int err = 1;
	FILE *in = fopen(filename, "rb");
	if(in == NULL)
	{
		fprintf(stderr, "[error] Unable to read %s. %s", filename, strerror(errno));
		return 0;
	}

	/* Retrieve data size */
	fseek(in, 0, SEEK_END);
	*dataSize = ftell(in);
	fseek(in, 0, SEEK_SET);
	*dataSize = *dataSize - ftell(in);

	/* Allocate data */
	tmp = (char*)realloc(*data, *dataSize + 1);
	if(tmp == NULL)
	{
		fclose(in);
		return 0;
	}
	*data = tmp;

	/* Read data */
	nRead = fread(*data, 1, *dataSize, in);
	(*data)[nRead] = '\0';
	if(nRead != *dataSize)
	{
		fprintf(stderr, "[error] Read %d out of %d bytes. %s", nRead, *dataSize, strerror(errno));
		err = 0;
	}

	fclose(in);
	return err;
}

/* Display shader compilation errors. */
void ShaderInfoLog(GLuint id)
{
	GLsizei maxLogLength, logLength;
	GLchar *log;

	glGetShaderiv(id, GL_INFO_LOG_LENGTH, (GLint*)&maxLogLength);

	log = (GLchar*)malloc(sizeof(GLchar) * maxLogLength);	

	glGetShaderInfoLog(id, maxLogLength, &logLength, log);
	
	fprintf(stderr,"[error] %s", log);

	free(log);
}

/* Display program link erros. */
void ProgramInfoLog(GLuint id)
{
	GLsizei maxLogLength, logLength;
	GLchar *log;

	glGetProgramiv(id, GL_INFO_LOG_LENGTH, (GLint*)&maxLogLength);	
	log = (GLchar*)malloc(sizeof(GLchar) * maxLogLength);
	
	glGetProgramInfoLog(id, maxLogLength, &logLength, log);

	fprintf(stderr,"[error] %s", log);
	
	free(log);
}

/* Check for compilation errors. 
 * Return GL_FALSE if the compilation failed. */
GLint CheckShaderCompilationErrors(GLuint i)
{
	GLint result;
	glGetShaderiv((i), GL_COMPILE_STATUS, &result);
	return result;
}

/* Check for link errors. 
 * Return GL_FALSE if the link failed. */
GLint CheckProgramLinkErrors(GLuint i) 
{
	GLint result;
	glGetProgramiv((i), GL_LINK_STATUS, &result);
	return result;
}

/* Main entry point. */
int main()
{
	GLenum err;
	int imageWidth, imageHeight, bpp;
	int windowWidth, windowHeight;
	unsigned char *imageData;
	int i, j;

	char *sourceData;
	size_t sourceSize;

	GLuint texId[TEXTURE_COUNT];
	GLuint fullscreenTriangleFBO, fullscreenTriangleVAO;
	GLuint program[PROGRAM_COUNT], shader;
	GLuint framebuffer[FRAMEBUFFER_COUNT];
	GLint scaleId;
	
	const char *filename = "data/mandrill.jpg";
	const char *vertexPositionAttributeName = "inPosition";

	struct {
		const char* vertexShaderFilename;
		const char* fragmentShaderFilename;
		
		const char* attributeName[4];
		const char* fragDataName[4];
	} programConfiguration[PROGRAM_COUNT] =
	{
		{ "basic.vert", "detail.frag", { "inPosition", 0, 0, 0 }, { "outCoarse", "outDetail", 0, 0 } },
		{ "basic.vert", "final.frag",  { "inPosition", 0, 0, 0 }, { "outData", 0, 0, 0 } },
	};

	const GLchar fullscreenTrianglePoints[6] =
	{
		-1,-1,
		 3,-1,
		-1, 3
	};

	const GLenum drawBuffers[2] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1 };

	/* Initialize stuff */
	err = 1;
	imageData = NULL;
	sourceData = NULL;
	fullscreenTriangleFBO = fullscreenTriangleVAO = 0;
	for(i=0; i<TEXTURE_COUNT; i++)
	{
		texId[i] = 0;
	}
	for(i=0; i<PROGRAM_COUNT; i++)
	{
		program[i] = 0;
	}
	for(i=0; i<FRAMEBUFFER_COUNT; i++)
	{
		framebuffer[i] = 0;
	}

	/* Load source image */
	imageData = stbi_load(filename, &imageWidth, &imageHeight, &bpp, 0);
	if(imageData == NULL)
	{
		fprintf(stderr, "[error] Unable to load %s : %s", filename, stbi_failure_reason());
		return 1;
	}

	/* Maximum image size is 2048x2048. */
	if((imageWidth > MAX_IMAGE_WIDTH) || (imageHeight > MAX_IMAGE_HEIGHT))
	{
		fprintf(stderr, "[error] The image is too big (%d, %d). The maximum supported image dimensions are (%d, %d).", 
			    imageWidth, imageHeight, MAX_IMAGE_WIDTH, MAX_IMAGE_HEIGHT);
		goto _error_;
	}

	/* Adjust window dimensions. */
	windowWidth = imageWidth;
	if(windowWidth > MAX_WINDOW_WIDTH)
	{
		windowWidth = MAX_WINDOW_WIDTH;
	}

	windowHeight = windowWidth * imageHeight / imageWidth;
	if(windowHeight > MAX_WINDOW_HEIGHT)
	{
		windowHeight = MAX_WINDOW_HEIGHT;
		windowWidth  = windowHeight * imageWidth / imageHeight;
	}
	
	/* Open window and set video mode. */
	if(glfwInit() != GL_TRUE)
	{
		fprintf(stderr, "[error] glfw failed to initialize.");
		goto _error_;
	}
	
	if(glfwOpenWindow(windowWidth, windowHeight, 8, 8, 8, 8, 0, 0, GLFW_WINDOW ) != GL_TRUE)
	{
		fprintf(stderr, "[error] Unable to open window (%d, %d).", windowWidth, windowHeight);
		goto _error_;
	}

	glfwSetWindowTitle("A Trous wavelet example");

	err = glewInit();
	if (GLEW_OK != err)
	{
		fprintf(stderr, "[error] Unable to initialize glew!");
		goto _error_;
	}

	/* glew version 1.5.6 did not load glBindFragDataLocation */
	if(__glewBindFragDataLocation == NULL)
	{
		__glewBindFragDataLocation = (PFNGLBINDFRAGDATALOCATIONPROC)glfwGetProcAddress("glBindFragDataLocation");
	}
	if(__glewBindFragDataLocation == NULL)
	{
		fprintf(stderr, "[error] Unable to retrieve a valid address for glBindFragDataLocation!");
		goto _error_;
	}

	/* The version of glfw I'm using generates an "invalid enumerant" OpenGL error. */
	/* If we don't "consume" it, it will show up at the next OpenGL error check. */
	err = glGetError();

	/* Upload image data to texture */
	glEnable(GL_TEXTURE_2D);
	glGenTextures(TEXTURE_COUNT, texId);

	/* This will be our detail textures */
	glBindTexture(GL_TEXTURE_2D_ARRAY, texId[0]);
		glTexImage3D(GL_TEXTURE_2D_ARRAY, 0, GL_RGB32F, imageWidth, imageHeight, LEVEL_MAX, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
		for(i=0; i<LEVEL_MAX; i++)
		{
			glTexSubImage3D(GL_TEXTURE_2D_ARRAY, 0, 0, 0, i, imageWidth, imageHeight, 1, GL_RGB, GL_UNSIGNED_BYTE, imageData);
		}
		glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_2D_ARRAY, 0);
	
	/* "Coarse" textures */
	for(i=1; i<TEXTURE_COUNT; i++)
	{
		glBindTexture(GL_TEXTURE_2D, texId[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, imageWidth, imageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, imageData);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}
	glBindTexture(GL_TEXTURE_2D, 0);
	
	/* Release image data */
	stbi_image_free(imageData);
	imageData = NULL;

	/* VBO for fullscreen triangle */
	glGenVertexArrays(1, &fullscreenTriangleVAO);
	glBindVertexArray(fullscreenTriangleVAO);
		glEnableVertexAttribArray(0);
		glGenBuffers(1, &fullscreenTriangleFBO);
		glBindBuffer(GL_ARRAY_BUFFER, fullscreenTriangleFBO);
			glBufferData(GL_ARRAY_BUFFER, 6*sizeof(GLchar), fullscreenTrianglePoints, GL_STATIC_DRAW);
			glVertexAttribPointer(0, 2, GL_BYTE, GL_FALSE, 0, 0);
	glBindVertexArray(0);
	
	/* Create programs */
	sourceData = NULL;
	sourceSize = 0;
	for(i=0; i<PROGRAM_COUNT; i++)
	{
		program[i] = glCreateProgram();

		/* Load and compile vertex shader. */
		LoadFile(programConfiguration[i].vertexShaderFilename, &sourceData, &sourceSize);
		shader = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(shader, 1, &sourceData, NULL);
		glCompileShader(shader);
		err = CheckShaderCompilationErrors(shader);
		if(!err)
		{
			fprintf(stderr, "[error] Shader compilation failed :", programConfiguration[i].vertexShaderFilename);
			ShaderInfoLog(shader);
			goto _error_;
		}
		glAttachShader(program[i], shader);

		/* Load and compile fragment shader. */
		LoadFile(programConfiguration[i].fragmentShaderFilename, &sourceData, &sourceSize);
		shader = glCreateShader(GL_FRAGMENT_SHADER);		
		glShaderSource(shader, 1, &sourceData, NULL);		
		glCompileShader(shader);
		err = CheckShaderCompilationErrors(shader);
		if(!err)
		{
			fprintf(stderr, "[error] Shader compilation failed :", programConfiguration[i].fragmentShaderFilename);
			ShaderInfoLog(shader);
			goto _error_;
		}
		glAttachShader(program[i], shader);

		/* Set input and output. */
		for(j=0; (j<4) && programConfiguration[i].attributeName[j]; j++)
		{
			glBindAttribLocation(program[i], j, programConfiguration[i].attributeName[j]);
		}
			
		for(j=0; (j<4) && programConfiguration[i].fragDataName[j]; j++)
		{
			glBindFragDataLocation(program[i], j, programConfiguration[i].fragDataName[j]);
		}
		glLinkProgram(program[i]);
		err = CheckProgramLinkErrors(program[i]);
		if(!err)
		{
			ProgramInfoLog(program[i]);
			goto _error_;
		}
	}
	if(sourceData != NULL)
	{
		free(sourceData);
		sourceData = NULL;
	}

	/* Create framebuffer objects */
	glGenFramebuffers(FRAMEBUFFER_COUNT, framebuffer);
	for(i=0; i<LEVEL_MAX; i++)
	{
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[i]);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texId[(i&1)?2:1], 0 ); 
		glFramebufferTextureLayer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, texId[0], 0, i);
	}
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[i]);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texId[(i&1)?2:1], 0 ); 

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	/* */
	glUseProgram(program[0]);
		glUniform1i(glGetUniformLocation(program[0], "uTex"), 0);
		scaleId = glGetUniformLocation(program[0], "uScale");
	glUseProgram(0);

	glUseProgram(program[1]);
		glUniform1i(glGetUniformLocation(program[1], "uTex"), 0);
		glUniform1i(glGetUniformLocation(program[1], "uDetails"), 1);
	glUseProgram(0);

	/* Compute detail texture. */
	glBindVertexArray(fullscreenTriangleVAO);
	glViewport(0, 0, imageWidth, imageHeight);
	for(i=0; i<LEVEL_MAX; i++)
	{
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[i]);
		glDrawBuffers(2, drawBuffers);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, texId[(i&1)?1:2]);

		glUseProgram(program[0]);
			glUniform1i(scaleId, 1<<i);
			glDrawArrays(GL_TRIANGLES, 0, 3);
		glUseProgram(0);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer[i]);
	glDrawBuffer(GL_COLOR_ATTACHMENT0);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texId[(i&1)?1:2]);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D_ARRAY, texId[0]);

	glUseProgram(program[1]);
	glDrawArrays(GL_TRIANGLES, 0, 3);
	glUseProgram(0);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D_ARRAY, 0);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, 0);

	/* Main Loop */
	while((glfwGetKey(GLFW_KEY_ESC) != GLFW_PRESS) && glfwGetWindowParam(GLFW_OPENED))
	{
		glViewport(0, 0, windowWidth, windowHeight);
		
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		glDrawBuffer(GL_BACK);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, texId[(i&1)?2:1]);
		glBegin(GL_QUADS);
			glTexCoord2f(0,1);
			glVertex2f(-1.0,-1.0);
			glTexCoord2f(1,1);
			glVertex2f(1.0,-1.0);
			glTexCoord2f(1,0);
			glVertex2f(1.0,1.0);
			glTexCoord2f(0,0);
			glVertex2f(-1.0,1.0);
		glEnd();

		glfwSwapBuffers();
	}

	glBindVertexArray(0);

	for(i=0; i<PROGRAM_COUNT; i++)
	{
		glDeleteProgram(program[i]);
	}

	err = 0;

_error_:
	glDeleteFramebuffers(FRAMEBUFFER_COUNT, framebuffer);

	if(fullscreenTriangleVAO)
	{
		glDeleteVertexArrays(1, &fullscreenTriangleVAO);
	}

	if(fullscreenTriangleFBO)
	{
		glDeleteBuffers(1, &fullscreenTriangleFBO);
	}
	
	glDeleteTextures(TEXTURE_COUNT, texId);

	glfwCloseWindow();
	glfwTerminate();
	
	if(imageData != NULL)
	{
		stbi_image_free(imageData);
	}
	if(sourceData != NULL)
	{
		free(sourceData);
	}

	return 0;
}